/*
*	CaliperI2CTest.ino
*	Copyright (c) 2018 Jonathan Mackey

*	GNU license:
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*	Please maintain this license information along with authorship and copyright
*	notices in any redistribution of this code.
*
*	Serial IO commands:
*	z - zeros the caliper
*	s - puts the caliper to sleep (turns if off)
*	w - wakes up the caliper (turns it on)
*	v - displays the current caliper value
*
*	Also, changes to the caliper value will be displayed whenever they occur.
*/
#include <Arduino.h>
#include <Wire.h>
#include "Caliper.h"

#define CALIPER_I2C_ADDR	0x21	// ATTiny84A I2C address
#define CaliperDataChangedPin 2		// Any interrupt capable pin
Caliper	caliper(CALIPER_I2C_ADDR, CaliperDataChangedPin);

void DisplayCaliperValue(void);

/*********************************** setup ************************************/
void setup(void)
{
	Serial.begin(19200);
	caliper.begin();
	Serial.println("Starting");
}


/*********************************** loop *************************************/
void loop(void)
{
	if (Serial.available())
	{
		int thisByte = Serial.read();
		switch(thisByte)
		{
			case 'Z':
			case 'z':
				caliper.Zero();
				break;
			case 'S':
			case 's':
				caliper.Sleep();
				break;
			case 'W':
			case 'w':
				caliper.Wake();
				break;
			case 'V':
			case 'v':
				caliper.ReadRaw();
				DisplayCaliperValue();
				break;
			case 'd':
				Serial.print("gDataChanged = ");
				Serial.println(caliper.DataChanged() ? "true" : "false");
				Serial.print("CaliperDataChangedPin = ");
				Serial.println(digitalRead(CaliperDataChangedPin) ? "HIGH" : "LOW");
				break;
		}
	} else if (caliper.DataChanged())
	{
		caliper.ReadRaw();
		DisplayCaliperValue();
	}
}

/***************************** DisplayCaliperValue ****************************/
void DisplayCaliperValue(void)
{
	uint32_t	rawValue = caliper.RawValue();
	if (caliper.ValueIsValid())
	{
		char buffer[50];
		char*	endBuffPtr = UInt32ToBinaryStr(rawValue, buffer, 0x800000);	// display the last 6 nibbles
		*endBuffPtr = 0;
		uint8_t	dPlaces;
		float	value = caliper.GetValue(&dPlaces);
		Serial.print(buffer);
		Serial.print(" = ");
		Serial.print(value, dPlaces);
		Serial.println(caliper.IsMetric() ? "mm" : "in");
	} else
	{
		Serial.println("Wire request failed");
	}
}

/****************************** UInt32ToBinaryStr *****************************/
char* UInt32ToBinaryStr(
	int32_t		inValue,
	char*		inBuffer,
	uint32_t	inMask)
{
	uint32_t	nMask = inMask;
	char*	buffPtr = inBuffer;
	for (; nMask != 0; nMask >>= 1)
	{
		*(buffPtr++) = (inValue & nMask) ? '1':'0';
		if (nMask & 0xEEEEEEEF)
		{
			continue;
		}
		*(buffPtr++) = ' ';
	}
	*buffPtr = 0;
	return(buffPtr);
}

